import inspect
from typing import Optional

from gantry.exceptions import ClientNotInitialized
from gantry.query.client import GantryQuery
from gantry.query.core.distance import GantryDistance
from gantry.query.core.metric import GantryMetric

_Query: Optional[GantryQuery] = None


def _query_alias(f):
    doc: str = ""
    signature = inspect.signature(f)
    orig_doc: Optional[str] = None
    if hasattr(GantryQuery, f.__name__):
        doc = "Alias for :meth:`gantry.query.core.GantryQuery.{}`".format(f.__name__)
        orig_doc = inspect.getdoc(getattr(GantryQuery, f.__name__))
        signature = inspect.signature(getattr(GantryQuery, f.__name__))
    elif hasattr(GantryDistance, f.__name__):
        doc = "Alias for :meth:`gantry.query.core.GantryDistance.{}`".format(f.__name__)
        orig_doc = inspect.getdoc(getattr(GantryDistance, f.__name__))
        signature = inspect.signature(getattr(GantryDistance, f.__name__))
    elif hasattr(GantryMetric, f.__name__):
        doc = "Alias for :meth:`gantry.query.core.GantryMetric.{}`".format(f.__name__)
        orig_doc = inspect.getdoc(getattr(GantryMetric, f.__name__))
        signature = inspect.signature(getattr(GantryMetric, f.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])

    f.__doc__ = doc
    f.__signature__ = signature
    return f


def validate_init():
    if _Query is None:
        raise ClientNotInitialized()



import mock
import pandas as pd
import pytest
import responses
from responses import matchers

from gantry.query.core.metric import GantryMetric, _prepare_pandas_dataframe

from ..conftest import END_TIME, ORIGIN, START_TIME, to_tz_time


@pytest.fixture
def gantry_metric_obj(api_client_obj):
    return GantryMetric(api_client_obj)


@pytest.mark.parametrize("data_nodes_size", [0, 3, 5])
def test_metric_query_invalid_data_nodes(data_nodes_size, gantry_metric_obj, series_obj_factory):
    test_data_nodes = [series_obj_factory("str") for _ in range(data_nodes_size)]
    with pytest.raises(ValueError):
        gantry_metric_obj._metric_query(
            data_nodes=test_data_nodes,
            stat="this-is-not-used",
            num_points=1,
        )


@pytest.mark.parametrize("num_points", [-2, -1, 0])
def test_metric_query_invalid_num_points(num_points, gantry_metric_obj, series_obj_factory):
    series_obj_1 = series_obj_factory("str")
    series_obj_2 = series_obj_factory("str")
    with pytest.raises(ValueError):
        gantry_metric_obj._metric_query(
            data_nodes=[series_obj_1, series_obj_2],
            stat="this-is-not-used",
            num_points=num_points,
        )


def test_metric_query(gantry_metric_obj, series_obj_factory_with_filters_and_tags):
    series_obj_1 = series_obj_factory_with_filters_and_tags("float", [1, 2, 3], {"foo": "bar"})
    series_obj_2 = series_obj_factory_with_filters_and_tags("bool", [1, 2, 3], None)
    data = {
        "start_time": to_tz_time(START_TIME),
        "end_time": to_tz_time(END_TIME),
        "num_points": 10,
        "queries": {
            "query_label": {
                "query_type": "data_node_ids",
                "model_node_id": "ABCD1234",
                "stat": "stat",
                "stat_kwargs": {"foo": "bar"},
                "filters": [1, 2, 3],
                "start_time": to_tz_time(START_TIME),
                "end_time": to_tz_time(END_TIME),
                "data_node_ids": ["abc123", "abc123"],
            },
        },
    }
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/time_series/query".format(ORIGIN),
            json={"response": "ok", "data": {"query_label": {"points": 12345}}},
            match=[matchers.json_params_matcher(data)],
        )

        assert (
            gantry_metric_obj._metric_query(
                [series_obj_1, series_obj_2], stat="stat", num_points=10, stat_kwargs={"foo": "bar"}
            )
            == 12345
        )


def test_check_comparable_invalid_series_filters(
    gantry_metric_obj, series_obj_factory, api_client_obj, query_info_obj
):
    parent_dataframe_1 = mock.Mock(api_client=api_client_obj, query_info=query_info_obj)
    parent_dataframe_2 = mock.Mock(api_client=api_client_obj, query_info=query_info_obj)
    parent_dataframe_1.filters = ["some-filter"]
    parent_dataframe_2.filters = ["some-other-filter"]

    series_obj_1 = series_obj_factory("str", parent_dataframe=parent_dataframe_1)
    series_obj_2 = series_obj_factory("str", parent_dataframe=parent_dataframe_2)

    with pytest.raises(ValueError):
        gantry_metric_obj._metric_query(
            data_nodes=[series_obj_1, series_obj_2],
            stat="this-is-not-used",
        )


@pytest.mark.parametrize("dropna", [True, False])
def test_empty_results_returns_empty_df(dropna):
    df = _prepare_pandas_dataframe(stat="not-used", results=[], dropna=dropna)
    assert isinstance(df, pd.DataFrame)
    assert df.empty


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_accuracy_score(mock_query, num_points, dropna, gantry_metric_obj, series_obj_factory):
    stat = "accuracy_score"
    series_obj_1 = series_obj_factory("str")
    series_obj_2 = series_obj_factory("str")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.5}]
    result = gantry_metric_obj.accuracy_score(
        series_obj_1, series_obj_2, num_points=num_points, dropna=dropna
    )
    assert isinstance(result, pd.DataFrame)
    assert result.shape == (1, 1)
    assert result.columns.tolist() == [stat]
    assert result.index.name == "timestamp"
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.5
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
    )


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@pytest.mark.parametrize("multioutput", ["uniform_average", "raw_values", None])
@pytest.mark.parametrize("squared", [False, True])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_mean_squared_error(
    mock_query, squared, multioutput, num_points, dropna, gantry_metric_obj, series_obj_factory
):
    stat = "mean_squared_error"
    series_obj_1 = series_obj_factory("float")
    series_obj_2 = series_obj_factory("float")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.99}]
    result = gantry_metric_obj.mean_squared_error(
        series_obj_1,
        series_obj_2,
        multioutput=multioutput,
        squared=squared,
        num_points=num_points,
        dropna=dropna,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == [stat]
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.99
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
        stat_kwargs={"multioutput": multioutput, "squared": squared},
    )


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_confusion_matrix(mock_query, num_points, dropna, gantry_metric_obj, series_obj_factory):
    stat = "confusion_matrix"
    series_obj_1 = series_obj_factory("int")
    series_obj_2 = series_obj_factory("int")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": [[1, 2], [3, 4]]}]
    result = gantry_metric_obj.confusion_matrix(
        series_obj_1, series_obj_2, num_points=num_points, dropna=dropna
    )
    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == [stat]
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == [[1, 2], [3, 4]]
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
    )


@pytest.mark.parametrize("method", ["precision_score", "recall_score", "f1_score"])
@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_f1_score_precision_score_recall_score(
    mock_query, num_points, dropna, method, gantry_metric_obj, series_obj_factory
):
    stat = method
    average = "micro"
    series_obj_1 = series_obj_factory("int")
    series_obj_2 = series_obj_factory("int")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.5}]
    result = getattr(gantry_metric_obj, method)(
        series_obj_1, series_obj_2, average=average, num_points=num_points, dropna=dropna
    )
    assert isinstance(result, pd.DataFrame)
    assert result.shape == (1, 1)
    assert result.columns.tolist() == [stat]
    assert result.index.name == "timestamp"
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.5
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
        stat_kwargs={"average": average},
    )


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@pytest.mark.parametrize("multioutput", ["uniform_average", "raw_values", None])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_r2_score(
    mock_query, multioutput, num_points, dropna, gantry_metric_obj, series_obj_factory
):
    stat = "r2_score"
    series_obj_1 = series_obj_factory("float")
    series_obj_2 = series_obj_factory("float")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.99}]
    result = gantry_metric_obj.r2_score(
        series_obj_1,
        series_obj_2,
        multioutput=multioutput,
        num_points=num_points,
        dropna=dropna,
    )
    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == [stat]
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.99
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
        stat_kwargs={"multioutput": multioutput},
    )


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
def test_roc_auc_score(mock_query, num_points, dropna, gantry_metric_obj, series_obj_factory):
    stat = "roc_auc_score"
    series_obj_1 = series_obj_factory("float")
    series_obj_2 = series_obj_factory("bool")
    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.68}]
    result = gantry_metric_obj.roc_auc_score(
        series_obj_1, series_obj_2, num_points=num_points, dropna=dropna
    )
    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == [stat]
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.68
    mock_query.assert_called_once_with(
        [series_obj_1, series_obj_2],
        stat,
        num_points=num_points,
    )


@pytest.mark.parametrize("dropna", [False, True])
@pytest.mark.parametrize("num_points", [1, 2, 10])
@pytest.mark.parametrize(
    "func_name, stat",
    [
        ["percent_null", "percent_null"],
        ["percent_true", "percent_true"],
        ["percent_false", "percent_false"],
        ["percent_true_not_null", "percent_true_not_null"],
        ["percent_false_not_null", "percent_false_not_null"],
    ],
)
@mock.patch("gantry.query.core.metric.GantryMetric._metric_query")
@mock.patch("gantry.query.core.metric._prepare_pandas_dataframe")
@mock.patch("gantry.query.core.dataframe.GantrySeries.__getitem__")
def test_percent_metric(
    mock_getitem,
    mock_prepare_pandas_dataframe,
    mock_query,
    func_name,
    stat,
    num_points,
    dropna,
    gantry_metric_obj,
    series_obj_factory,
):
    # TODO -> decouple this test case in simpler test cases with
    # more explicit mocks and specific calls
    series_obj_1 = series_obj_factory("bool")
    series_obj_2 = series_obj_factory("bool")

    mock_prepare_pandas_dataframe.side_effect = _prepare_pandas_dataframe

    series_obj_1.notna = mock.Mock(return_value="foobar")
    mock_getitem.return_value = series_obj_2

    timestamp = "2022-03-20T00:00:00"
    mock_query.return_value = [{"time_stamp": timestamp, "value": 0.2}]
    result = getattr(gantry_metric_obj, func_name)(
        series_obj_1, num_points=num_points, dropna=dropna
    )
    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == [stat]
    result_score = result.loc[result.index == timestamp, stat].values[0]
    assert result_score == 0.2

    mock_prepare_pandas_dataframe.assert_called_with(
        stat, [{"time_stamp": "2022-03-20T00:00:00", "value": 0.2}], dropna=False
    )
    if dropna and func_name != "percent_null":
        mock_query.assert_called_once_with(
            [series_obj_2],
            stat,
            num_points=num_points,
        )
        mock_getitem.assert_called_with("foobar")
    else:
        mock_query.assert_called_once_with(
            [series_obj_1],
            stat,
            num_points=num_points,
        )
        series_obj_1.notna.assert_not_called()


@pytest.mark.parametrize(
    "func_name",
    ["percent_true", "percent_false", "percent_true_not_null", "percent_false_not_null"],
)
def test_percent_bool_invalid_data_type(func_name, gantry_metric_obj, series_obj_factory):
    test_data_node = series_obj_factory("str")
    with pytest.raises(NotImplementedError):
        getattr(gantry_metric_obj, func_name)(
            data_node=test_data_node,
            stat="this-is-not-used",
            num_points=1,
        )



from enum import Enum


class BatchType(str, Enum):
    RECORD = "RECORD"
    FEEDBACK = "FEEDBACK"
    PREDICTION = "PREDICTION"


class UploadFileType(str, Enum):
    CSV_WITH_HEADERS = "CSV_WITH_HEADERS"
    EVENTS = "EVENTS"


class Delimiter(Enum):
    COMMA = ","


CHUNK_SIZE = 20 * 1024 * 1024  # File upload chunk size 20MB


class ScheduleFrequency(str, Enum):
    EVERY_HOUR = "1 hour"
    EVERY_2_HOURS = "2 hours"
    EVERY_4_HOURS = "4 hours"
    EVERY_8_HOURS = "8 hours"
    EVERY_12_HOURS = "12 hours"
    EVERYDAY = "1 day"


class ScheduleType(str, Enum):
    INCREMENTAL_APPEND = "INCREMENTAL_APPEND"
    FULL_REFRESH_APPEND = "FULL_REFRESH_APPEND"



import functools
from datetime import datetime
from inspect import getcallargs, getfullargspec
from typing import Dict, List, Optional, Tuple, Union, cast

import dateparser
import isodate
import pandas as pd

from gantry.api_client import APIClient


def same_size(*params: str, allow_empty: bool = False):
    """
    Use this decorator to prevent running methods over iterables that are different sizes.
    For example, if 'some_method(s1, s2, s3)' need s1/s2/s3 to be same size then
    it can be decorated with '@same_size("s1", "s2", "s3")'.

    Args:
        *params (List[str]): Name of parameters that will be passed to the decorated
            method
        allow_empty (bool, defaults to False): In case False, empty Parameters
            (with length of 0) will fail the check.

    """

    def decorator(method):
        fullargs = getfullargspec(method).args
        if not all(map(lambda x: x in fullargs, params)):
            raise ValueError(
                f"Decorator on {method} was badly parametrized {params}. This shouldn't happen. "
                "Parameters for this decorator should be a subset "
                "of the methods allowed parameters."
            )

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            provided_params = getcallargs(method, *args, **kwargs)
            sizes = {}
            for param in params:
                size = len(provided_params[param])
                if not allow_empty and len(provided_params[param]) == 0:
                    raise ValueError(f"Parameter '{param}' has size 0. This is not allowed.")
                sizes[param] = size
            if len(set(sizes.values())) != 1:
                raise ValueError(
                    f"Size mismatch on inputs: {sizes}. "
                    f"All GantrySeries passed need to be same size."
                )
            return method(*args, **kwargs)

        return wrapper

    return decorator


def runs_on(*types):
    def decorator(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if self.dtype not in types:
                raise ValueError(
                    f"'{self.name}' has type '{self.datatype}'. "
                    f"'{method.__name__}' support {types} only"
                )
            return method(self, *args, **kwargs)

        return wrapper

    return decorator


def get_application_views(
    api_client: APIClient,
    application: str,
    version=Optional[Union[str, int]],
    environment=Optional[str],
) -> list:
    version_param = version or get_last_application_version(
        api_client=api_client, application=application
    )

    response = api_client.request(
        "GET",
        "/api/v1/applications/" + application + "/views",
        params={"version": version_param},
        raise_for_status=True,
    )

    views = response["data"]

    if environment:
        views = list(filter(lambda view: view["tag_filters"].get("env") == environment, views))

    return views


def get_last_application_version(
    api_client: APIClient,
    application: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> str:
    query_params: dict[str, Union[int, str, datetime]] = {}
    if start_time:
        query_params["start_time"] = start_time
    if end_time:
        query_params["end_time"] = end_time
    metadata_response = api_client.request(
        "GET", "/api/v1/models/" + application + "/schemas", query_params, raise_for_status=True
    )
    if metadata_response.get("data", {}).get("version") is None:
        raise RuntimeError("Unkown error: couldn't fetch last application version from API.")

    return metadata_response["data"]["version"]


def get_application_node_id(
    api_client: APIClient,
    application: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    version: Optional[Union[str, int]] = None,
) -> str:
    query_params: dict[str, Union[int, str, datetime]] = {}
    if start_time:
        query_params["start_time"] = start_time
    if end_time:
        query_params["end_time"] = end_time
    if version:
        query_params["version"] = version

    metadata_response = api_client.request(
        "GET", "/api/v1/models/" + application + "/schemas", query_params, raise_for_status=True
    )
    if metadata_response.get("data", {}).get("id") is None:
        raise RuntimeError("Unkown error: couldn't fetch node id from API.")

    return metadata_response["data"]["id"]


def get_start_end_time_from_view(view: dict) -> Tuple[datetime, datetime]:
    if view.get("start_time"):
        start_time = cast(datetime, dateparser.parse(view["start_time"]))
        end_time = cast(datetime, dateparser.parse(view["end_time"]))
    elif view.get("duration"):
        start_time = datetime.utcnow()
        end_time = start_time - isodate.parse_duration(view["duration"])
    else:
        raise ValueError(
            "View has invalid time configuration. Either start_time/end_time"
            " or duration need to be set"
        )

    return start_time, end_time


def _build_empty_df_with_schema(schema: Dict[str, str]) -> pd.DataFrame:
    cols = [item[0] for item in schema.items()]
    ret = pd.DataFrame(columns=cols)
    for col, dtype in schema.items():
        try:
            ret[col] = ret[col].astype(dtype)
        except TypeError:
            ret[col] = ret[col].astype("str")
    return ret


def _build_batch_id_filter(batch_id: str) -> dict:
    return {
        "feature_name": "batch_id",
        "string_query": batch_id,
        "dtype": "reserved_field",
    }


def _get_application_batches(
    api_client: APIClient,
    application: str,
    version: Optional[Union[str, int]] = None,
) -> List[dict]:
    version_param = version or get_last_application_version(
        api_client=api_client, application=application
    )
    node_id = get_application_node_id(api_client, application, version=version_param)
    response = api_client.request(
        "GET",
        "/api/v1/applications/" + node_id + "/batches",
        raise_for_status=True,
    )

    return response["data"]["batches"]



# Methods that are not implemented yet (raise NotImplementedError)
# are not exposed here.

from gantry.query import distance, metric  # noqa: F401
from gantry.query.main import (  # noqa: F401
    create_view,
    init,
    list_application_batches,
    list_application_environments,
    list_application_versions,
    list_application_views,
    list_applications,
    print_application_info,
    query,
)

__all__ = [
    "init",
    "query",
    "list_applications",
    "list_application_environments",
    "list_application_versions",
    "list_application_views",
    "list_application_batches",
    "print_application_info",
    "create_view",
]



import uuid

import mock
import responses
from click.testing import CliRunner
from responses import matchers

from gantry.cli.secret import create, delete, list

SECRET_NAME = "test_secret"
UNITTEST_HOST = "http://unittest"
SECRET_INFO = {
    "id": str(uuid.uuid4()),
    "secret_name": SECRET_NAME,  # type: ignore
    "organization_id": str(uuid.uuid4()),
    "secret_type": "AWS",
    "created_at": "2022-12-15T00:05:23",
}


@mock.patch("gantry.cli.secret.APIClient.request")
def test_create_secret_failure(mock_api_client, datadir):
    mock_api_client.return_value = {"response": "error", "error": "some error"}

    runner = CliRunner()
    cli_args = [
        "--name",
        SECRET_NAME,
        "--secret-type",
        "aws",
        "--secret-file",
        str(datadir / "secret.json"),
    ]
    result = runner.invoke(
        create,
        cli_args,
        env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
    )
    assert result.exit_code == 1
    assert f"Fail to create secret {SECRET_NAME}: some error" in result.output


def test_create_secret(datadir):
    runner = CliRunner()
    cli_args = [
        "--name",
        SECRET_NAME,
        "--secret-type",
        "aws",
        "--secret-file",
        str(datadir / "secret.json"),
    ]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/customer-resources/secrets",
            json={
                "response": "ok",
                "data": SECRET_INFO,
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "secret_name": SECRET_NAME,
                        "secret_type": "AWS",
                        "secret_content": {
                            "access_key": "test-access-key",
                            "secret_key": "test-secret-key",
                        },
                    }
                )
            ],
        )

        result = runner.invoke(
            create,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert "--> Secret has been created." in result.output


def test_list_secrets():
    runner = CliRunner()
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/secrets",
            json={"response": "ok", "data": [SECRET_INFO]},
            headers={"Content-Type": "application/json"},
        )

        result = runner.invoke(
            list,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert f"{SECRET_INFO}" in result.output


def test_delete_secret():
    runner = CliRunner()
    cli_args = ["--secret-id", SECRET_INFO["id"]]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.DELETE,
            f"{UNITTEST_HOST}/api/v1/customer-resources/secrets/{SECRET_INFO['id']}",
            json={
                "response": "ok",
            },
            headers={"Content-Type": "application/json"},
        )

        result = runner.invoke(
            delete,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert f"--> Secret {SECRET_INFO['id']} has been deleted." in result.output



import mock
import pytest
from freezegun import freeze_time

from gantry.exceptions import GantryLoggingDataTypeError
from gantry.logger import event_builder

from .conftest import CURRENT_TIME, CURRENT_TIME_STR, SOME_TIME, SOME_TIME_STR


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize(
    ["sort", "timestamps", "record_count", "expected"],
    [
        (False, [3, 4, 5, 1, 2], 10, [(0, 3), (1, 4), (2, 5), (3, 1), (4, 2)]),
        (True, [3, 4, 5, 1, 2], 10, [(3, 1), (4, 2), (0, 3), (1, 4), (2, 5)]),
        (False, ["b", "a", "c"], 100, [(0, "b"), (1, "a"), (2, "c")]),
        (True, ["b", "a", "c"], 100, [(1, "a"), (0, "b"), (2, "c")]),
        (
            True,
            None,
            4,
            [
                (0, CURRENT_TIME),
                (1, CURRENT_TIME),
                (2, CURRENT_TIME),
                (3, CURRENT_TIME),
            ],
        ),
        (
            False,
            None,
            4,
            [
                (0, CURRENT_TIME),
                (1, CURRENT_TIME),
                (2, CURRENT_TIME),
                (3, CURRENT_TIME),
            ],
        ),
    ],
)
def test_create_timestamp_idx(sort, timestamps, record_count, expected):
    assert list(event_builder._create_timestamp_idx(sort, timestamps, record_count)) == expected


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize("feedback_version", [None, 5, "1.2.3"])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_build_feedback_event(mock_uuid, feedback_version, timestamp):
    mock_uuid.return_value = "123abc"
    assert event_builder._build_feedback_event(
        application="barbaz",
        join_key="1234567890",
        feedback={"b": 20},
        feedback_version=feedback_version,
        timestamp=timestamp,
        batch_id="ABCD1234",
    ) == {
        "batch_id": "ABCD1234",
        "event_id": "123abc",
        "feedback": {"b": 20},
        "feedback_id": "1234567890",
        "log_timestamp": CURRENT_TIME_STR,
        "metadata": {
            "feedback_version": feedback_version,
            "func_name": "barbaz",
        },
        "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
    }


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_build_prediction_event(mock_uuid, timestamp):
    mock_uuid.return_value = "123abc"
    assert event_builder._build_prediction_event(
        inputs={"some-input": "a-input"},
        outputs={"some-output": "a-output"},
        application="barbaz",
        join_key="1234567890",
        version="1.2.3",
        ignore_inputs=["some-other-input"],
        batch_id="ABCD1234",
        tags={"this-is": "a-tag"},
        custom_timestamp=timestamp,
    ) == {
        "batch_id": "ABCD1234",
        "event_id": "123abc",
        "feedback_id": "1234567890",
        "inputs": {"some-input": "a-input"},
        "log_timestamp": CURRENT_TIME_STR,
        "metadata": {
            "feedback_keys": None,
            "version": "1.2.3",
            "ignore_inputs": ["some-other-input"],
            "func_name": "barbaz",
        },
        "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
        "outputs": {"some-output": "a-output"},
        "tags": {"this-is": "a-tag"},
    }


def test_build_prediction_event_invalid_input():
    with pytest.raises(GantryLoggingDataTypeError):
        event_builder._build_prediction_event(
            inputs={"some-invalid-input": ["invalid", {"invalid": True}]},
            outputs={"some-output": "a-output"},
            application="barbaz",
            join_key="1234567890",
            version="1.2.3",
            ignore_inputs=[],
            batch_id="ABCD1234",
            tags={"this-is": "a-tag"},
            custom_timestamp=SOME_TIME,
        )



import datetime
import json
import logging
import os
import uuid
from io import StringIO
from pathlib import Path

import datasets
import pandas as pd
import pytest
import responses
from datasets import Sequence, Value, builder
from mock import patch
from responses import matchers

from gantry.api_client import APIClient
from gantry.dataset.constants import (
    DATASET_MANIFEST_FILE,
    DATASET_STASH_FILE,
    DELETED_FILES,
    MODIFIED_FILES,
    NEW_FILES,
    STASH_FOLDER,
)
from gantry.dataset.gantry_dataset import DatasetFileInfo, GantryDataset
from gantry.exceptions import (
    DataSchemaMismatchError,
    DatasetCommitNotFoundError,
    DatasetHeadOutOfDateException,
    DatasetNotFoundError,
    GantryException,
    GantryRequestException,
    NoTabularDataError,
)
from gantry.query.core.dataframe import GantryDataFrame
from gantry.utils import get_files_checksum, list_all_files

from .conftest import (
    AWS_REGION,
    BUCKET_NAME,
    COMMIT_MSG,
    CONF_OBJ_KEY,
    CSV_OBJ_KEY,
    DATASET_ID,
    DATASET_NAME,
    HOST,
    IMG_OBJ_KEY,
    MANIFEST_OBJ_KEY,
    MANIFEST_VERSION_ID,
    README_OBJ_KEY,
    S3_PREFIX,
)


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture(scope="function")
def gantry_dataset_obj(test_api_client, datadir):
    return GantryDataset(
        api_client=test_api_client,
        dataset_name=DATASET_NAME,
        dataset_id=DATASET_ID,
        bucket_name=BUCKET_NAME,
        aws_region=AWS_REGION,
        dataset_s3_prefix=S3_PREFIX,
        workspace=datadir,
    )


def test_list_commits(gantry_dataset_obj, commit_history):
    """
    Test get dataset commits
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        commits = gantry_dataset_obj.list_versions()
        assert len(commits) == 2


def test_get_commit(gantry_dataset_obj, commit_history):
    """
    Test get commit information
    """
    expected_commit = commit_history[1]
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits/{expected_commit['id']}",
            json={
                "response": "ok",
                "data": expected_commit,
            },
            headers={"Content-Type": "application/json"},
        )
        commit = gantry_dataset_obj._get_commit(expected_commit["id"])
        assert commit == expected_commit


@pytest.mark.parametrize(
    ("status_code", "expected_error"),
    [(404, DatasetCommitNotFoundError), (500, GantryRequestException)],
)
def test_get_commit_error(gantry_dataset_obj, commit_history, status_code, expected_error):
    commit_id = commit_history[0]["id"]
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits/{commit_id}",
            status=status_code,
            headers={"Content-Type": "application/json"},
        )
        with pytest.raises(expected_error):
            gantry_dataset_obj._get_commit(commit_id)


def test_get_latest_commit(gantry_dataset_obj, commit_history):
    """
    Test get latest commit
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        latest_commit = gantry_dataset_obj._get_latest_commit()
        assert latest_commit["id"] == "8ee0f6d5-c84c-473a-931a-5148b2e704d7"
        assert latest_commit["is_latest_commit"]


def test_get_diff(gantry_dataset_obj):
    """
    Test get local edit diff
    """
    # overwrite the dataset name
    gantry_dataset_obj.dataset_name = "show_diff_dataset"

    diff = gantry_dataset_obj.get_diff()

    assert set(diff[NEW_FILES]) == {"artifacts/kitten.png", "tabular_manifests/feedback.csv"}
    assert diff[MODIFIED_FILES] == ["dataset_config.yaml"]
    assert diff[DELETED_FILES] == ["README.md"]


def test_get_diff_not_exist(gantry_dataset_obj):
    """
    Test get diff without download the dataset
    """
    # overwrite the dataset name
    gantry_dataset_obj.dataset_name = "dataset_not_exist"
    with pytest.raises(DatasetNotFoundError):
        gantry_dataset_obj.get_diff()


def test_create_commit_without_download(gantry_dataset_obj):
    """
    Test create commit when no dataset downloaded locally
    """
    # overwrite the dataset name
    gantry_dataset_obj.dataset_name = "dataset_not_exist"
    with pytest.raises(DatasetNotFoundError):
        gantry_dataset_obj.push_version("test commit")


@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._get_manifest_presign_url",
    return_value="s3://manifest/presign/url",
)
@patch("gantry.dataset.gantry_dataset.GantryDataset._get_tabular_files", return_value=[])
@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._extract_feature_schema_from_s3", return_value={}
)
def test_pull_huggingface_dataset_by_version_no_schema(
    mock_get_presign_url,
    mock_get_tabular_files,
    mock_extract_feature_schema_from_s3,
    gantry_dataset_obj,
):
    """
    Test pull huggingface dataset by version
    """
    with pytest.raises(DataSchemaMismatchError):
        gantry_dataset_obj.pull_huggingface_dataset_by_version(
            hf_dataset_name="hf_test",
            version_id="1.0.0",
        )


@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._get_manifest_presign_url",
    return_value="s3://manifest/presign/url",
)
@patch("gantry.dataset.gantry_dataset.GantryDataset._get_tabular_files", return_value=[])
@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._extract_feature_schema_from_s3",
    return_value={"feature": "schema"},
)
def test_pull_huggingface_dataset_by_version_no_data(
    mock_get_presign_url,
    mock_get_tabular_files,
    mock_extract_feature_schema_from_s3,
    gantry_dataset_obj,
):
    """
    Test pull huggingface dataset by version
    """
    with pytest.raises(NoTabularDataError):
        gantry_dataset_obj.pull_huggingface_dataset_by_version(
            hf_dataset_name="hf_test",
            version_id="1.0.0",
        )


@pytest.mark.parametrize("correct_schema, streaming", [(True, True), (True, False), (False, False)])
@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._get_manifest_presign_url",
    return_value="s3://manifest/presign/url",
)
@patch("gantry.dataset.gantry_dataset.GantryDataset._get_tabular_files")
@patch("gantry.dataset.gantry_dataset.GantryDataset._extract_feature_schema_from_s3")
@patch("gantry.dataset.gantry_dataset.GantryDataset._generate_get_presign_url")
def test_pull_huggingface_dataset_by_version(
    mock_generate_get_presign_url,
    mock_extract_feature_schema_from_s3,
    mock_get_tabular_files,
    mock_get_manifest_presign_url,
    correct_schema,
    streaming,
    gantry_dataset_obj,
    datadir,
):
    """
    Test pull huggingface dataset by version
    """

    if correct_schema:
        mock_extract_feature_schema_from_s3.return_value = {
            "inputs.credit_history": Value(dtype="string", id=None),
            "inputs.credit_scores": Sequence(
                feature=Value(dtype="float64", id=None), length=-1, id=None
            ),
            "inputs.description": Value(dtype="string", id=None),
            "inputs.image_uri": Value(dtype="string", id=None),
            "inputs.income": Value(dtype="int64", id=None),
            "inputs.loan_id": Value(dtype="string", id=None),
            "inputs.prediction": Value(dtype="bool", id=None),
            "inputs.timestamp": Value(dtype="timestamp[ns, tz=UTC]", id=None),
            "outputs.output": Value(dtype="float64", id=None),
            "feedback.loan_repaid": Value(dtype="bool", id=None),
            "__time": Value(dtype="timestamp[ns, tz=UTC]", id=None),
        }
    else:
        mock_extract_feature_schema_from_s3.return_value = {
            "inputs.credit_history": Value(dtype="string", id=None),
            "inputs.prediction": Value(dtype="bool", id=None),
            "inputs.timestamp": Value(dtype="timestamp[ns, tz=UTC]", id=None),
            "outputs.output": Value(dtype="float64", id=None),
            "feedback.loan_repaid": Value(dtype="bool", id=None),
            "__time": Value(dtype="timestamp[ns, tz=UTC]", id=None),
        }

    def side_effect(**kwargs):
        return {f.file_name: f.url for f in kwargs["batch"]}

    mock_get_tabular_files.return_value = [
        DatasetFileInfo(
            file_name="mock_tabular_file_1.csv",
            # for unit test we use local file path instead of s3 url
            url=str(datadir / "hf-test" / "tabular_manifests" / "hf_dataset_1.csv"),
            version_id="mock_version_id",
        ),
        DatasetFileInfo(
            file_name="mock_tabular_file_2.csv",
            url=str(datadir / "hf-test" / "tabular_manifests" / "hf_dataset_2.csv"),
            version_id="mock_version_id",
        ),
    ]
    mock_generate_get_presign_url.side_effect = side_effect

    if correct_schema:
        hf_dataset = gantry_dataset_obj.pull_huggingface_dataset_by_version(
            version_id="mock_version_id", hf_dataset_name="hf_test", streaming=streaming
        )
        if not streaming:
            assert len(hf_dataset) == 40
        else:
            assert len(list(hf_dataset)) == 40
            hf_dataset.__class__ = datasets.iterable_dataset.IterableDataset

        assert hf_dataset.features == {
            "inputs.credit_history": Value(dtype="string", id=None),
            "inputs.credit_scores": Sequence(
                feature=Value(dtype="float64", id=None), length=-1, id=None
            ),
            "inputs.description": Value(dtype="string", id=None),
            "inputs.image_uri": Value(dtype="string", id=None),
            "inputs.income": Value(dtype="int64", id=None),
            "inputs.loan_id": Value(dtype="string", id=None),
            "inputs.prediction": Value(dtype="bool", id=None),
            "inputs.timestamp": Value(dtype="timestamp[ns, tz=UTC]", id=None),
            "outputs.output": Value(dtype="float64", id=None),
            "feedback.loan_repaid": Value(dtype="bool", id=None),
            "__time": Value(dtype="timestamp[ns, tz=UTC]", id=None),
        }
        assert hf_dataset.builder_name == "hf_test"
    else:
        with pytest.raises(DataSchemaMismatchError):
            hf_dataset = gantry_dataset_obj.pull_huggingface_dataset_by_version(
                version_id="mock_version_id", hf_dataset_name="hf_test"
            )


def test_get_tabular_files(gantry_dataset_obj, datadir):
    with responses.RequestsMock() as resp:
        MOCK_URL = "http://mock-url"
        manifest_file = datadir / "hf-test" / ".dataset_metadata" / ".gantry_manifest.jsonl"
        resp.add(responses.GET, MOCK_URL, body=manifest_file.read_text())
        tabular_files = gantry_dataset_obj._get_tabular_files(MOCK_URL)
        assert len(tabular_files) == 2
        assert tabular_files[0].file_name.endswith(".csv")
        assert tabular_files[0].file_name.endswith(".csv")


@patch("gantry.dataset.gantry_dataset.GantryDataset._get_obj_key", return_value="mock_obj")
@patch(
    "gantry.dataset.gantry_dataset.GantryDataset._generate_get_presign_url",
    return_value={"mock_obj": "http://mock-config-url"},
)
def test_extract_feature_schema_from_s3(
    mock_generate_get_presign_url, mock_get_obj_key, gantry_dataset_obj, datadir
):
    MANIFEST_URL = "http://mock-manifest-url"
    with responses.RequestsMock() as resp:
        manifest_file = datadir / "hf-test" / ".dataset_metadata" / ".gantry_manifest.jsonl"
        resp.add(responses.GET, MANIFEST_URL, body=manifest_file.read_text())
        config_file = datadir / "hf-test" / "dataset_config.yaml"
        resp.add(responses.GET, "http://mock-config-url", body=config_file.read_text())
        features = gantry_dataset_obj._extract_feature_schema_from_s3(MANIFEST_URL)
        assert len(features) == 11


@pytest.mark.parametrize("use_latest_commit", [True, False])
@patch("gantry.dataset.gantry_dataset.GantryDataset._get_commit")
@patch("gantry.dataset.gantry_dataset.GantryDataset._get_latest_commit")
@patch("gantry.dataset.gantry_dataset.GantryDataset._generate_get_presign_url")
def test_get_manifest_presign_url(
    mock_generate_get_presign_url,
    mock_get_latest_commit,
    mock_get_commit,
    use_latest_commit,
    commit_history,
    gantry_dataset_obj,
):
    mock_get_latest_commit.return_value = commit_history[0]
    mock_get_commit.return_value = commit_history[1]

    def side_effect(args):
        return {gantry_dataset_obj._get_obj_key(args[0].url): f"{args[0].url}-presign"}

    mock_generate_get_presign_url.side_effect = side_effect
    if use_latest_commit:
        presign_url = gantry_dataset_obj._get_manifest_presign_url()
    else:
        presign_url = gantry_dataset_obj._get_manifest_presign_url(
            version_id=commit_history[1]["id"]
        )
    assert presign_url.endswith("presign")


@pytest.mark.parametrize(
    ["dataset_name", "hf_name", "expected_name", "streaming"],
    [
        ("hf-test", "hf_name", "hf_name", False),
        ("hf-test", None, "hftest", False),
        ("hf-test", "hf_name", "hf_name", True),
        ("hf-test", None, "hftest", True),
    ],
)
def test_get_huggingface_dataset(
    dataset_name, hf_name, expected_name, streaming, gantry_dataset_obj
):
    """
    Test get local edit diff
    """
    # overwrite the dataset name
    gantry_dataset_obj.dataset_name = dataset_name

    hf_dataset = gantry_dataset_obj.get_huggingface_dataset(
        hf_dataset_name=hf_name, streaming=streaming
    )

    if not streaming:
        assert len(hf_dataset) == 40
    else:
        len(list(hf_dataset)) == 40
        assert hf_dataset.__class__ == datasets.iterable_dataset.IterableDataset

    assert hf_dataset.features == {
        "inputs.credit_history": Value(dtype="string", id=None),
        "inputs.credit_scores": Sequence(
            feature=Value(dtype="float64", id=None), length=-1, id=None
        ),
        "inputs.description": Value(dtype="string", id=None),
        "inputs.image_uri": Value(dtype="string", id=None),
        "inputs.income": Value(dtype="int64", id=None),
        "inputs.loan_id": Value(dtype="string", id=None),
        "inputs.prediction": Value(dtype="bool", id=None),
        "inputs.timestamp": Value(dtype="timestamp[ns, tz=UTC]", id=None),
        "outputs.output": Value(dtype="float64", id=None),
        "feedback.loan_repaid": Value(dtype="bool", id=None),
        "__time": Value(dtype="timestamp[ns, tz=UTC]", id=None),
    }
    assert hf_dataset.builder_name == expected_name


@pytest.mark.parametrize(
    ["dataset_name", "hf_name", "load_from_remote"],
    [
        ("12-123", None, True),
        ("hf_test", "123", True),
        ("12-123", None, False),
        ("hf_test", "123", False),
    ],
)
def test_get_huggingface_dataset_error(dataset_name, hf_name, load_from_remote, gantry_dataset_obj):
    gantry_dataset_obj.dataset_name = dataset_name
    if load_from_remote:
        hf_dataset = gantry_dataset_obj.pull_huggingface_dataset_by_version(hf_dataset_name=hf_name)
    else:
        hf_dataset = gantry_dataset_obj.get_huggingface_dataset(hf_dataset_name=hf_name)

    assert hf_dataset is None


def test_get_huggingface_dataset_no_folder(gantry_dataset_obj):
    """
    Test error when no tabular manifests folder
    """
    gantry_dataset_obj.dataset_name = "no_tabular_manifest_dataset"

    with pytest.raises(NoTabularDataError):
        gantry_dataset_obj.get_huggingface_dataset()


def test_get_huggingface_dataset_mismatch_schema(gantry_dataset_obj):
    """
    Test error when no tabular manifests folder
    """
    gantry_dataset_obj.dataset_name = "mismatch_config_dataset"

    with pytest.raises(DataSchemaMismatchError):
        gantry_dataset_obj.get_huggingface_dataset()


def test_get_huggingface_dataset_bad_data_type_schema(gantry_dataset_obj):
    """
    Test error when no tabular manifests folder
    """
    gantry_dataset_obj.dataset_name = "bad_data_config_dataset"

    with pytest.raises(DataSchemaMismatchError) as e:
        gantry_dataset_obj.get_huggingface_dataset()

    assert isinstance(e.value.__cause__, builder.DatasetGenerationError)


def setup_resp_for_deleted_check(resp, is_deleted=False, dataset_name=DATASET_NAME):
    resp.add(
        resp.GET,
        f"{HOST}/api/v1/datasets/{dataset_name}",
        json={
            "response": "ok",
            "data": {
                "id": DATASET_ID,
                "name": dataset_name,
                "organization_id": "fake-org-id",
                "bucket_name": BUCKET_NAME,
                "disabled": is_deleted,
            },
        },
        headers={"Content-Type": "application/json"},
    )


def setup_resp_for_head_check(resp, latest_commit, dataset_id=DATASET_ID):
    resp.add(
        resp.GET,
        f"{HOST}/api/v1/datasets/{dataset_id}/commits",
        json={
            "response": "ok",
            "data": [latest_commit],
        },
        headers={"Content-Type": "application/json"},
    )


def setup_resp_for_create_commit(
    datadir, resp, create_commit_resp_json, commit_id, create_commit_status_code=200
):
    # fake urls for uploading files
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/presign/putobject",
        json={
            "response": "ok",
            "data": {
                MANIFEST_OBJ_KEY: f"{HOST}/fake_presigned_url",
                CONF_OBJ_KEY: f"{HOST}/fake_presigned_url",
                README_OBJ_KEY: f"{HOST}/fake_presigned_url",
                IMG_OBJ_KEY: f"{HOST}/fake_presigned_url",
                CSV_OBJ_KEY: f"{HOST}/fake_presigned_url",
            },
        },
        headers={"Content-Type": "application/json"},
    )
    resp.add(
        resp.PUT,
        f"{HOST}/fake_presigned_url",
        json={"response": "ok"},
        headers={"Content-Type": "application/json", "x-amz-version-id": MANIFEST_VERSION_ID},
    )

    with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
        old_commit = json.load(f)

    # for the actual commit creation request
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
        json=create_commit_resp_json,
        headers={"Content-Type": "application/json"},
        status=create_commit_status_code,
        match=[
            matchers.json_params_matcher(
                {
                    "message": COMMIT_MSG,
                    "metadata_s3_file_version": MANIFEST_VERSION_ID,
                    "parent_commit_id": old_commit["id"],
                    "commit_id": commit_id,
                }
            )
        ],
    )


def setup_resp_for_sync(
    resp,
    commit_id,
    commit_json,
    manifest_file,
    dataset_id=DATASET_ID,
    do_fake_presigned=True,
):
    # getting commit info
    if commit_id is not None:  # if commit_id is specified
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{dataset_id}/commits/{commit_id}",
            json={
                "response": "ok",
                "data": commit_json,
            },
            headers={"Content-Type": "application/json"},
        )
    else:  # if we're syncing to latest
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{dataset_id}/commits",
            json={
                "response": "ok",
                "data": [commit_json],
            },
            headers={"Content-Type": "application/json"},
        )

    # getting the fake urls for the objects
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{dataset_id}/presign/getobject",
        json={
            "response": "ok",
            "data": {
                MANIFEST_OBJ_KEY: f"{HOST}/fake_manifest",
                CONF_OBJ_KEY: f"{HOST}/fake_presigned_url",
                README_OBJ_KEY: f"{HOST}/fake_presigned_url",
            },
        },
        headers={"Content-Type": "application/json"},
    )

    # adding responses for the fake download urls
    req_kwargs = {
        "stream": True,
    }
    resp.add(
        resp.GET,
        f"{HOST}/fake_manifest",
        body=manifest_file.read(),
        match=[matchers.request_kwargs_matcher(req_kwargs)],
    )
    if do_fake_presigned:
        resp.add(
            resp.GET,
            f"{HOST}/fake_presigned_url",
            body="fake_response",
            match=[matchers.request_kwargs_matcher(req_kwargs)],
        )


def setup_resp_for_add_file(resp, commit_message, filename, parent_commit_id=None):
    file_dir = "artifacts"
    if len(filename) > 4 and filename[-4:] == ".csv":
        file_dir = "tabular_manifests"

    file_url = f"org_id/datasets/v1/{DATASET_NAME}/test-uuid-1/{file_dir}/{filename}"

    # for uploading the added file and the manifest file
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/presign/putobject",
        json={
            "response": "ok",
            "data": {
                file_url: f"{HOST}/fake_presigned_url_0",
                MANIFEST_OBJ_KEY: f"{HOST}/fake_presigned_url_1",
            },
        },
        headers={"Content-Type": "application/json"},
    )
    # for downloading the manifest file
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/presign/getobject",
        json={
            "response": "ok",
            "data": {
                MANIFEST_OBJ_KEY: f"{HOST}/fake_presigned_url_2",
            },
        },
        headers={"Content-Type": "application/json"},
    )

    # fake presigned urls for downloading/uploading files
    resp.add(
        resp.PUT,
        f"{HOST}/fake_presigned_url_0",
        json={"response": "ok"},
        headers={"x-amz-version-id": "fake_version_id"},
    )
    resp.add(
        resp.PUT,
        f"{HOST}/fake_presigned_url_1",
        json={"response": "ok"},
        headers={"x-amz-version-id": MANIFEST_VERSION_ID},
    )
    resp.add(
        resp.GET,
        f"{HOST}/fake_presigned_url_2",
        body="test-string",
    )

    # for getting the latest commit
    if not parent_commit_id:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": [{"id": "4569ef0b-ff96-4110-a606-8d9d54fab382"}],
            },
            headers={"Content-Type": "application/json"},
        )

    # for the actual commit creation request
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
        json={
            "data": {
                "created_at": "Wed, 07 Oct 2022 06:11:02 GMT",
                "created_by": "db459d6d-c83b-496d-b659-e48bca971156",
                "dataset_id": DATASET_ID,
                "id": "add-file-commit-id",
                "is_latest_commit": True,
                "message": commit_message,
                "metadata_s3_file_version": MANIFEST_VERSION_ID,
                "parent_commit": "8ee0f6d5-c84c-473a-931a-5148b2e704d7",
            },
        },
        headers={"Content-Type": "application/json"},
        status=200,
        match=[
            matchers.json_params_matcher(
                {
                    "message": commit_message,
                    "metadata_s3_file_version": MANIFEST_VERSION_ID,
                    "parent_commit_id": parent_commit_id or "4569ef0b-ff96-4110-a606-8d9d54fab382",
                    "commit_id": "test-uuid-1",
                }
            )
        ],
    )


@patch("uuid.uuid4", return_value="test-uuid-1")
def test_create_commit(mock_uuid, datadir, gantry_dataset_obj, commit_history):
    """
    Test create commit succeed
    """
    with responses.RequestsMock() as resp:
        setup_resp_for_create_commit(
            datadir, resp, {"response": "ok", "data": commit_history[0]}, "test-uuid-1"
        )
        setup_resp_for_deleted_check(resp)
        with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
            setup_resp_for_head_check(resp, json.load(f))  # we are on latest commit

        gantry_dataset_obj.push_version(COMMIT_MSG)

        diff = gantry_dataset_obj.get_diff()

        # verify no diff after a successful commit
        assert not diff[NEW_FILES]
        assert not diff[MODIFIED_FILES]
        assert not diff[DELETED_FILES]

        with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
            assert json.load(f) == commit_history[0]  # verify HEAD has been updated


@patch("uuid.uuid4", return_value="test-uuid-1")
def test_create_commit_failure(mock_uuid, datadir, gantry_dataset_obj, commit_history):
    """
    Test create commit failure, in this case we detect the head is not up to date and stop
    prematurely. This will happen during a race condition when another user committed a change and
    the local copy has not been synced subsequently.
    """
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_head_check(resp, commit_history[0])  # we are NOT on latest commit

        with pytest.raises(DatasetHeadOutOfDateException):
            gantry_dataset_obj.push_version(COMMIT_MSG)

        diff = gantry_dataset_obj.get_diff()

        # since commit failed the local diff will be the same
        assert set(diff[NEW_FILES]) == {"artifacts/kitten.png", "tabular_manifests/feedback.csv"}
        assert diff[MODIFIED_FILES] == ["dataset_config.yaml"]
        assert diff[DELETED_FILES] == ["README.md"]


def test_create_commit_nochange(datadir, gantry_dataset_obj):
    """
    Test create commit without any local change
    """
    dataset_name = "unittest_dataset"

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp, dataset_name=dataset_name)

        gantry_dataset_obj.dataset_name = dataset_name
        with open(datadir / f"{dataset_name}/.dataset_metadata/HEAD") as f:
            old_commit = json.load(f)

        new_commit = gantry_dataset_obj.push_version("new commit")
        assert old_commit == new_commit


def create_mock_manifest(commit_history):
    gantry_manifest = StringIO()
    config_file_info = {
        "file_name": "dataset_config.yaml",
        "url": f"s3://test-bucket/{CONF_OBJ_KEY}",
        "sha256": "unmatched_sha256",
        "version_id": "random_vid_for_config",
    }

    gantry_manifest.write(f"{json.dumps(config_file_info)}\n")

    readme_file_info = {
        "file_name": "README.md",
        "url": f"s3://test-bucket/{README_OBJ_KEY}",
        "sha256": "unmatched_sha256",
        "version_id": "random_vid_for_readme",
    }

    gantry_manifest.write(f"{json.dumps(readme_file_info)}\n")
    gantry_manifest.seek(0)
    commit_json = commit_history[1]

    return gantry_manifest, commit_json


@pytest.mark.parametrize("commit_id", [None, "eb2e7242-3340-4edf-8366-90d4fce897ce"])
def test_pull_dataset(commit_id, datadir, gantry_dataset_obj, commit_history):
    """
    Test sync local data set
    1. To latest commit
    2. To a specific commit
    """
    manifest_file, commit_json = create_mock_manifest(commit_history)

    gantry_dataset_obj.dataset_name = "show_diff_dataset"

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_sync(resp, commit_id, commit_json, manifest_file)

        commit = gantry_dataset_obj.pull(commit_id, forced=True)
        assert commit == gantry_dataset_obj._prune_commit_info(commit_json)
        with open(datadir / "show_diff_dataset/dataset_config.yaml", "r") as f:
            assert f.read() == "fake_response"

        with open(datadir / "show_diff_dataset/README.md", "r") as f:
            assert f.read() == "fake_response"

        with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
            assert json.load(f) == commit_json

        assert not list_all_files(Path(datadir / "show_diff_dataset/tabular_manifests"))
        assert not list_all_files(Path(datadir / "show_diff_dataset/artifacts"))


def test_pull_local_changes(datadir, gantry_dataset_obj):
    """
    Test trying to pull when local changes exist without forced=True
    """
    with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
        original_head = json.load(f)

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        commit = gantry_dataset_obj.pull()

        assert commit is None

        with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
            assert original_head == json.load(f)


def test_to_jsonl():
    file_info = DatasetFileInfo(
        file_name="test_file",
        url="s3://test-bucket/key",
        version_id="version_id",
        sha256="mock_sha256",
    )
    assert (
        file_info.to_jsonl()
        == '{"file_name": "test_file", "url": "s3://test-bucket/key", "version_id": "version_id", \
"sha256": "mock_sha256"}\n'
    )


@pytest.mark.parametrize(
    ["file_name", "url", "version_id", "sha256"],
    [
        (None, "s3://test-bucket/key", "version_id", "mock_sha256"),
        ("test_file", None, "version_id", "mock_sha256"),
        ("test_file", "s3://test-bucket/key", None, "mock_sha256"),
        ("test_file", "s3://test-bucket/key", "version_id", None),
    ],
)
def test_to_jsonl_value_error(file_name, url, version_id, sha256):
    file_info = DatasetFileInfo(
        file_name=file_name,
        url=url,
        version_id=version_id,
        sha256=sha256,
    )
    with pytest.raises(ValueError):
        file_info.to_jsonl()


def test_rollback(datadir, gantry_dataset_obj, commit_history):
    """
    Test rollback to previous commit
    """
    manifest_file, _ = create_mock_manifest(commit_history)
    original_commit = commit_history[1]
    new_commit = commit_history[0]

    rollback_commit = {
        "created_at": "Wed, 05 Oct 2022 12:38:05 GMT",
        "created_by": new_commit["created_by"],
        "dataset_id": DATASET_ID,
        "id": "311851c4-88d3-474f-9e06-13380d72e508",
        "is_latest_commit": True,
        "message": f"Rollback dataset to commit: {original_commit['id']}",
        "metadata_s3_file_version": original_commit["metadata_s3_file_version"],
        "parent_commit": new_commit["id"],
    }

    # sync to local
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_sync(resp, None, commit_history[0], manifest_file)

        gantry_dataset_obj.pull(forced=True)

    # try the rollback
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits/{original_commit['id']}",
            json={
                "response": "ok",
                "data": original_commit,
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={"response": "ok", "data": rollback_commit},
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "message": f"Rollback dataset to version: {original_commit['id']}",
                        "metadata_s3_file_version": original_commit["metadata_s3_file_version"],
                        "parent_commit_id": new_commit["id"],
                    }
                )
            ],
        )
        setup_resp_for_sync(
            resp,
            rollback_commit["id"],
            rollback_commit,
            manifest_file,
            do_fake_presigned=False,
        )

        gantry_dataset_obj.rollback(original_commit["id"], forced=True)

        with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
            assert json.load(f) == rollback_commit  # verify rollback has been successful


def test_rollback_local_changes(datadir, gantry_dataset_obj):
    """
    Test trying to pull when local changes exist without forced=True
    """
    with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
        original_head = json.load(f)

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        commit = gantry_dataset_obj.rollback("fake-uuid")

        assert commit is None

        with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
            assert original_head == json.load(f)


@pytest.mark.parametrize("function", ["push_version", "pull", "rollback"])
def test_function_on_deleted_dataset(datadir, gantry_dataset_obj, function):
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp, is_deleted=True)

        with pytest.raises(GantryException):
            if function == "push_version":
                gantry_dataset_obj.push_version("fake commit message")
            elif function == "pull":
                gantry_dataset_obj.pull()
            elif function == "rollback":
                gantry_dataset_obj.rollback("eb2e7242-3340-4edf-8366-90d4fce897ce")


def test_delete_dataset(gantry_dataset_obj):
    with responses.RequestsMock() as resp:
        resp.add(
            resp.DELETE,
            f"{HOST}/api/v1/datasets/{DATASET_ID}",
            json={
                "response": "ok",
            },
            headers={"Content-Type": "application/json"},
        )

        gantry_dataset_obj.delete()


@patch("uuid.uuid4", return_value="test-uuid-1")
@pytest.mark.parametrize(
    ("filename", "expected_filename"),
    [
        (None, "test_file.txt"),
        ("myfile.txt", "myfile.txt"),
    ],
)
@pytest.mark.parametrize("commit_message", [None, "test commit message"])
@pytest.mark.parametrize("read_method", ["rt", "rb"])
def test_add_file(
    mock_uuid, datadir, gantry_dataset_obj, filename, expected_filename, commit_message, read_method
):
    expected_commit_message = commit_message or f"Uploaded {expected_filename} to dataset"

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_add_file(
            resp,
            expected_commit_message,
            expected_filename,
        )

        gantry_dataset_obj.push_file(
            open(datadir / "test_file.txt", read_method), filename, commit_message
        )


@patch("uuid.uuid4", return_value="test-uuid-1")
@pytest.mark.parametrize(
    ("real_filename", "provided_filename", "expected_filename"),
    [
        ("test_file.csv", None, "test_file.csv"),
        ("test_file.csv", "mycsv.csv", "mycsv.csv"),
        ("bad_filename.txt", "mycsv.csv", "mycsv.csv"),
    ],
)
@pytest.mark.parametrize("commit_message", [None, "test commit message"])
def test_add_tabular_file(
    mock_uuid,
    gantry_dataset_obj,
    real_filename,
    provided_filename,
    expected_filename,
    commit_message,
):
    expected_commit_message = commit_message or f"Uploaded {expected_filename} to dataset"

    csvfile = StringIO()
    pd.DataFrame({"A": [1, 2], "B": [3, 4]}).to_csv(csvfile)
    csvfile.seek(0)
    csvfile.name = real_filename

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_add_file(
            resp,
            expected_commit_message,
            expected_filename,
        )

        csvfile.seek(0)
        gantry_dataset_obj.push_tabular_file(csvfile, provided_filename, commit_message)


@pytest.mark.parametrize(
    ("real_filename", "provided_filename"),
    [
        ("bad_filename.txt", None),
        ("test_file.csv", "mybadfilename.txt"),
    ],
)
def test_add_tabular_file_bad_name(gantry_dataset_obj, real_filename, provided_filename):
    csvfile = StringIO()
    pd.DataFrame({"A": [1, 2], "B": [3, 4]}).to_csv(csvfile)
    csvfile.name = real_filename

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)

        with pytest.raises(ValueError):
            gantry_dataset_obj.push_tabular_file(csvfile, provided_filename)


@patch("uuid.uuid4", return_value="test-uuid-1")
@pytest.mark.parametrize(
    ("filename", "expected_filename"),
    [
        (None, "dataframe-34dbe12d2b6000ce7875188f426df47d.csv"),
        ("myfile1", "myfile1.csv"),
        ("myfile2.csv", "myfile2.csv"),
    ],
)
@pytest.mark.parametrize("commit_message", [None, "test commit message"])
def test_add_dataframe(mock_uuid, gantry_dataset_obj, filename, expected_filename, commit_message):
    expected_commit_message = commit_message or f"Uploaded {expected_filename} to dataset"

    df = pd.DataFrame({"A": [100, 101], "B": [200, 202]})

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_add_file(
            resp,
            expected_commit_message,
            expected_filename,
        )

        gantry_dataset_obj.push_dataframe(df, filename, commit_message)


@patch("uuid.uuid4", return_value="uuid-uuid-uuid")
@pytest.mark.parametrize("filename", ["", "mock_filename"])
def test_push_gantry_dataframe(mock_uuid, gantry_dataset_obj, test_api_client, filename):
    start_time = datetime.datetime.now()
    end_time = datetime.datetime.now()
    final_filename_prefixes = {
        "": f"GantryDataAutoSave__{uuid.uuid4()}_{start_time.strftime('%Y-%m-%d_%H:%M')}_to_{end_time.strftime('%Y-%m-%d_%H:%M')}",  # noqa: E501
        "mock_filename": "mock_filename",
    }
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/foobar/schemas",
            json={
                "response": "ok",
                "data": {"id": "app_model_node_id"},
            },
            headers={"Content-Type": "application/json"},
        )
        gantrydf = GantryDataFrame(
            api_client=test_api_client,
            application="foobar",
            version="1.2.3",
            env="pytest",
            start_time=start_time,
            end_time=end_time,
            filters=[],
        )
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/raw_data",
            json={
                "response": "ok",
                "data": {
                    "id": "default-version-id",
                    "message": "default-message",
                    "created_at": "creation-timestamp",
                    "created_by": "gantry",
                    "is_latest_commit": True,
                },
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "model_id": "app_model_node_id",
                        "raw_data_query_parameters": {
                            "filters": gantrydf.filters,
                            "tags": gantrydf.tags,
                            "include_join_id": True,
                            "hidden_columns": None,
                            "start_time": str(start_time.isoformat()) + "Z",
                            "end_time": str(end_time.isoformat()) + "Z",
                        },
                        "raw_data_csv_prefix": final_filename_prefixes[filename],
                    }
                )
            ],
        )
        gantry_dataset_obj.push_gantry_dataframe(gantrydf, filename)


def update_manifest_with_checksums(datadir, dataset):
    """
    Updates the gantry manifest so that there is no local dif
    """
    local_files = get_files_checksum(datadir / dataset.dataset_name)

    with open(dataset._get_file_path(DATASET_MANIFEST_FILE), "w") as gantry_manifest:
        for file_name, checksum in local_files.items():
            file_info = {
                "file_name": file_name,
                "url": f"{S3_PREFIX}/{DATASET_NAME}/{file_name}",
                "sha256": checksum,
                "version_id": "01234567-89ab-cdef-0123-456789abcdef",
            }
            gantry_manifest.write(f"{json.dumps(file_info)}\n")


def test_stash_files(gantry_dataset_obj, datadir, commit_history):
    file_content = [{"test": "json"}, {"tomato": "fruit"}, {"oh no": "my file"}]
    file_names = ["newfile.json", "modfile.json", "delfile.json"]
    modified_file_content = {"tomato": "vegetable"}

    # prepare files
    with open(datadir / f"{DATASET_NAME}/{file_names[1]}", "w+") as f:
        json.dump(file_content[1], f)
    with open(datadir / f"{DATASET_NAME}/{file_names[2]}", "w+") as f:
        json.dump(file_content[2], f)
    # set dataset diff to be none
    update_manifest_with_checksums(datadir, gantry_dataset_obj)

    # modify files
    with open(datadir / f"{DATASET_NAME}/{file_names[0]}", "w+") as f:
        json.dump(file_content[0], f)
    with open(datadir / f"{DATASET_NAME}/{file_names[1]}", "w+") as f:
        json.dump(modified_file_content, f)
    os.remove(datadir / f"{DATASET_NAME}/{file_names[2]}")

    # check that pull fails
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        assert gantry_dataset_obj.pull() is None

    # stash
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets/{gantry_dataset_obj._dataset_id}/presign/getobject",
            json={
                "response": "ok",
                "data": {
                    f"datasets/v1/{DATASET_NAME}/{file_names[1]}": f"{HOST}/file1",
                    f"datasets/v1/{DATASET_NAME}/{file_names[2]}": f"{HOST}/file2",
                },
            },
            headers={"Content-Type": "application/json"},
        )

        # adding responses for the fake download urls
        req_kwargs = {
            "stream": True,
        }
        resp.add(
            resp.GET,
            f"{HOST}/file1",
            body=json.dumps(file_content[1]),
            match=[matchers.request_kwargs_matcher(req_kwargs)],
        )
        resp.add(
            resp.GET,
            f"{HOST}/file2",
            body=json.dumps(file_content[2]),
            match=[matchers.request_kwargs_matcher(req_kwargs)],
        )

        gantry_dataset_obj.stash()

    # check that stash was successful
    # check that relevant files are stashed
    with open(datadir / f"{DATASET_NAME}/{STASH_FOLDER}/{file_names[0]}") as f:
        assert json.load(f) == file_content[0]
    with open(datadir / f"{DATASET_NAME}/{STASH_FOLDER}/{file_names[1]}") as f:
        assert json.load(f) == modified_file_content
    # check that the stash file is updated
    with open(datadir / f"{DATASET_NAME}/{DATASET_STASH_FILE}") as f:
        assert json.load(f) == {
            DELETED_FILES: [file_names[2]],
            MODIFIED_FILES: [file_names[1]],
            NEW_FILES: [file_names[0]],
        }
    # check that the local files are in the correct state
    assert not os.path.exists(datadir / f"{DATASET_NAME}/{file_names[0]}")
    for i in [1, 2]:
        with open(datadir / f"{DATASET_NAME}/{file_names[i]}") as f:
            assert json.load(f) == file_content[i]

    # check that pull now succeeds
    manifest_file, commit_json = create_mock_manifest(commit_history)
    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        setup_resp_for_sync(resp, None, commit_json, manifest_file)
        assert gantry_dataset_obj.pull() == gantry_dataset_obj._prune_commit_info(commit_json)

    # add back file 2 since pulling deletes it
    with open(datadir / f"{DATASET_NAME}/{file_names[2]}", "w+") as f:
        json.dump(file_content[2], f)

    # retrieve stash
    gantry_dataset_obj.restore()

    # check that restore was successful
    # check that local files are in the correct state
    with open(datadir / f"{DATASET_NAME}/{file_names[0]}") as f:
        assert json.load(f) == file_content[0]
    with open(datadir / f"{DATASET_NAME}/{file_names[1]}") as f:
        assert json.load(f) == modified_file_content
    assert not os.path.exists(datadir / f"{DATASET_NAME}/{file_names[2]}")
    # check stash files are gone
    assert not os.path.exists(datadir / f"{DATASET_NAME}/{DATASET_STASH_FILE}")
    for i in [0, 1, 2]:
        assert not os.path.exists(datadir / f"{DATASET_NAME}/{STASH_FOLDER}/{file_names[i]}")


def test_double_stash(gantry_dataset_obj, datadir):
    file_content = [{"test": "json"}, {"test": "different json"}]
    file_names = ["newfile.json", "otherfile.json"]

    # set dataset diff to be none
    update_manifest_with_checksums(datadir, gantry_dataset_obj)

    # modify files and stash
    with open(datadir / f"{DATASET_NAME}/{file_names[0]}", "w+") as f:
        json.dump(file_content[0], f)
    gantry_dataset_obj.stash()

    # check that stash was successful
    with open(datadir / f"{DATASET_NAME}/{STASH_FOLDER}/{file_names[0]}") as f:
        assert json.load(f) == file_content[0]
    with open(datadir / f"{DATASET_NAME}/{DATASET_STASH_FILE}") as f:
        assert json.load(f) == {NEW_FILES: [file_names[0]], MODIFIED_FILES: [], DELETED_FILES: []}
    assert not os.path.exists(datadir / f"{DATASET_NAME}/{file_names[0]}")

    # modify files and stash again
    with open(datadir / f"{DATASET_NAME}/{file_names[1]}", "w+") as f:
        json.dump(file_content[1], f)
    gantry_dataset_obj.stash()

    # check that stash was again successful
    with open(datadir / f"{DATASET_NAME}/{STASH_FOLDER}/{file_names[1]}") as f:
        assert json.load(f) == file_content[1]
    with open(datadir / f"{DATASET_NAME}/{DATASET_STASH_FILE}") as f:
        assert json.load(f) == {NEW_FILES: [file_names[1]], MODIFIED_FILES: [], DELETED_FILES: []}
    assert not os.path.exists(datadir / f"{DATASET_NAME}/{file_names[1]}")


def test_load_nonexistent_stash(gantry_dataset_obj, caplog):
    with caplog.at_level(logging.WARNING):
        gantry_dataset_obj.restore()
    assert "No stashed files were found!" in caplog.text


@pytest.mark.parametrize(("function"), [("push_file"), ("push_tabular_file"), ("push_dataframe")])
def test_add_item_bad_parent(gantry_dataset_obj, function, commit_history):
    parent_commit_id = uuid.UUID("01234567-89ab-cdef-0123-456789abcdef")

    if function == "push_dataframe":
        input = pd.DataFrame()
    else:
        input = StringIO()
        input.name = "file.csv"

    with pytest.raises(DatasetHeadOutOfDateException):
        with responses.RequestsMock() as resp:
            setup_resp_for_deleted_check(resp)
            resp.add(
                resp.GET,
                f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
                json={
                    "response": "ok",
                    "data": commit_history,
                },
                headers={"Content-Type": "application/json"},
            )

            if function == "push_file":
                gantry_dataset_obj.push_file(input, parent_version_id=parent_commit_id)
            elif function == "push_tabular_file":
                gantry_dataset_obj.push_tabular_file(input, parent_version_id=parent_commit_id)
            elif function == "push_dataframe":
                gantry_dataset_obj.push_dataframe(input, parent_version_id=parent_commit_id)


@patch("uuid.uuid4", return_value="test-uuid-1")
@pytest.mark.parametrize(("function"), [("push_file"), ("push_tabular_file"), ("push_dataframe")])
def test_add_item_good_parent(mock_uuid, gantry_dataset_obj, function, commit_history):
    parent_commit_id = commit_history[0]["id"]
    filename = "test-file.csv"

    if function == "push_dataframe":
        input = pd.DataFrame()
    else:
        input = StringIO()
        input.name = filename

    with responses.RequestsMock() as resp:
        setup_resp_for_deleted_check(resp)
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        setup_resp_for_add_file(
            resp, f"Uploaded {filename} to dataset", filename, parent_commit_id=parent_commit_id
        )

        if function == "push_file":
            new_commit = gantry_dataset_obj.push_file(
                input, filename=filename, parent_version_id=parent_commit_id
            )
        elif function == "push_tabular_file":
            new_commit = gantry_dataset_obj.push_tabular_file(
                input, filename=filename, parent_version_id=parent_commit_id
            )
        elif function == "push_dataframe":
            new_commit = gantry_dataset_obj.push_dataframe(
                input, filename=filename, parent_version_id=parent_commit_id
            )

        assert new_commit["version_id"] == "add-file-commit-id"
        assert new_commit["dataset"] == gantry_dataset_obj.dataset_name
        assert new_commit["message"] == f"Uploaded {filename} to dataset"



import datetime
import logging
import uuid
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from gantry.exceptions import GantryLoggingException
from gantry.utils import check_event_time_in_future
from gantry.validators import validate_logged_field_datatype

logger = logging.getLogger(__name__)


def _build_prediction_and_feedback_events(
    application: str,
    timestamp_idx: Iterable,
    inputs: List[dict],
    outputs: Union[List[dict], List[Any]],
    feedbacks: List[dict],
    join_keys: List[str],
    version: Optional[Union[str, int]] = None,
    ignore_inputs: Optional[List[str]] = None,
    tags: Optional[List[Dict[str, str]]] = None,
    run_id: Optional[str] = None,
):
    """Build prediction and feedback events for log_records"""
    events = []
    for idx, timestamp in timestamp_idx:
        event = {}

        event.update(
            _build_prediction_event(
                inputs[idx],
                outputs[idx],
                application,
                join_keys[idx],
                version,
                ignore_inputs,
                custom_timestamp=timestamp,
                tags=tags[idx] if tags else None,
                batch_id=run_id,
            )
        )

        feedback_event = _build_feedback_event(
            application,
            join_keys[idx],
            feedbacks[idx],
            version,  # type: ignore
            timestamp,
            tags=tags[idx] if tags else None,
            batch_id=run_id,
        )

        # Build an event with prediction data
        # and feedback data, now that backend
        # supports it.
        event["metadata"].update(feedback_event.pop("metadata"))
        event.update(feedback_event)

        events.append(event)

    return events


def _create_timestamp_idx(
    sort_on_timestamp: bool,
    timestamps: Optional[Iterable[Any]],
    record_count: int,
) -> Iterable:
    """
    This function is used to take an index of timestamps, possibly None, and a length,
    and turn it into a mapping from the current timestamp to the original index of the
    corresponding value. The goal is to send all data to Gantry in timestamp order to
    minimize summarization window computing overhead by packing events of a single
    window into a single task.
    """
    if timestamps is not None:
        timestamp_idx: Iterable[Tuple[int, Any]] = enumerate(timestamps)
        if sort_on_timestamp:
            timestamp_idx = sorted(timestamp_idx, key=lambda el: el[1])
    else:
        timestamp = datetime.datetime.utcnow()
        timestamp_idx = ((i, timestamp) for i in range(record_count))

    return timestamp_idx


def _enrich_events_with_batch_id(events, batch_id):
    for event in events:
        event["batch_id"] = batch_id


def _build_feedback_event(
    application: str,
    join_key: str,
    feedback: dict,
    feedback_version: Optional[Union[str, int]] = None,
    timestamp: Optional[datetime.datetime] = None,
    batch_id: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
) -> dict:
    """
    Create a feedback event record for logging.
    """
    metadata: Dict[str, Any] = {}
    metadata.update({"func_name": application, "feedback_version": feedback_version})

    log_time = datetime.datetime.utcnow()
    event_time = timestamp if timestamp else log_time
    if check_event_time_in_future(event_time):
        raise GantryLoggingException(
            "Cannot log events from the future. "
            f"Event timestep is {event_time}, but current time is {log_time}. "
            "Please check your event times and re-submit all."
        )
    log_time_str, event_time_str = log_time.isoformat(), event_time.isoformat()

    validate_logged_field_datatype(feedback, paths=["feedback"])
    validate_logged_field_datatype(event_time_str, paths=["timestamp"])

    return {
        "event_id": uuid.uuid4(),
        "timestamp": event_time_str,
        "log_timestamp": log_time_str,
        "metadata": metadata,
        "feedback": feedback,
        "feedback_id": join_key,
        "batch_id": batch_id,
    }


def _build_prediction_events(
    application: str,
    inputs: List[dict],
    outputs: Union[List[dict], List[Any]],
    timestamp_idx: Iterable,
    version: Optional[Union[str, int]],
    join_keys: List[str],
    ignore_inputs: Optional[List[str]] = None,
    batch_id: Optional[str] = None,
    tags: Optional[List[Dict[str, str]]] = None,
):
    events = []
    for idx, timestamp in timestamp_idx:

        events.append(
            _build_prediction_event(
                inputs[idx],
                outputs[idx],
                application,
                join_keys[idx],
                version,
                ignore_inputs,
                batch_id=batch_id,
                custom_timestamp=timestamp,
                tags=tags[idx] if tags else None,
            )
        )

    return events


def _build_prediction_event(
    inputs: Dict,
    outputs: Any,
    application: str,
    join_key: str,
    version: Optional[Union[int, str]],
    ignore_inputs: Optional[List[str]],
    batch_id: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    custom_timestamp: Optional[datetime.datetime] = None,
):
    metadata = {}
    metadata.update(
        {
            "func_name": application,
            "version": version,
            "feedback_keys": None,  # TODO -> see if this can be removed
            "ignore_inputs": ignore_inputs,
        }
    )
    inputs = dict(inputs)  # make a copy so we can pop keys safely

    if ignore_inputs:
        for ignore in ignore_inputs:
            inputs.pop(ignore, None)

    log_time = datetime.datetime.utcnow()
    event_time = custom_timestamp if custom_timestamp else log_time
    if check_event_time_in_future(event_time):
        raise GantryLoggingException(
            "Cannot log events from the future. "
            f"Event timestep is {event_time}, but current time is {log_time}. "
            "Please check your event times and re-submit all."
        )
    log_time_str, event_time_str = log_time.isoformat(), event_time.isoformat()

    validate_logged_field_datatype(inputs, paths=["inputs"])
    validate_logged_field_datatype(outputs, paths=["outputs"])

    return {
        "event_id": uuid.uuid4(),
        "log_timestamp": log_time_str,
        "timestamp": event_time_str,
        "metadata": metadata,
        "inputs": inputs,
        "outputs": outputs,
        "feedback_id": join_key,
        "tags": tags,
        "batch_id": batch_id,
    }



import functools
import inspect
import os
from typing import Optional

from gantry.api_client import APIClient
from gantry.applications.client import ApplicationClient
from gantry.const import PROD_API_URL
from gantry.exceptions import ClientNotInitialized

_APPLICATION_CLIENT: Optional[ApplicationClient] = None


def _application_client_alias(f):
    doc = "Alias for :meth:`gantry.applications.client.ApplicationClient.{}`".format(f.__name__)
    signature = inspect.signature(getattr(ApplicationClient, f.__name__))
    orig_doc = inspect.getdoc(getattr(ApplicationClient, f.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])

    f.__doc__ = doc
    f.__signature__ = signature

    # This decorator also checks that the _CLIENT
    # has been initialized
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if _APPLICATION_CLIENT is None:
            raise ClientNotInitialized()

        return f(*args, **kwargs)

    return wrapper


def _init(api_key: Optional[str] = None):
    global _APPLICATION_CLIENT
    backend = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)
    api_client = APIClient(backend, api_key)
    _APPLICATION_CLIENT = ApplicationClient(api_client)  # type: ignore


@_application_client_alias
def create_application(*args, **kwargs):
    return _APPLICATION_CLIENT.create_application(*args, **kwargs)


@_application_client_alias
def get_application(*args, **kwargs):
    return _APPLICATION_CLIENT.get_application(*args, **kwargs)


@_application_client_alias
def archive_application(*args, **kwargs):
    return _APPLICATION_CLIENT.archive_application(*args, **kwargs)


@_application_client_alias
def delete_application(*args, **kwargs):
    return _APPLICATION_CLIENT.delete_application(*args, **kwargs)



import datetime
from typing import Optional, Union


class TimeWindow:
    def __init__(
        self, start_time: Union[str, datetime.datetime], end_time: Union[str, datetime.datetime]
    ):
        self.start_time = start_time
        self.end_time = end_time


class RelativeTimeWindow(TimeWindow):
    def __init__(
        self,
        window_length: datetime.timedelta,
        offset: Optional[datetime.timedelta] = None,
    ):
        self.window_length = window_length
        self.offset = offset
        if self.offset is None:
            self.offset = datetime.timedelta()
        end_time: datetime.datetime = (
            datetime.datetime.utcnow() - self.offset if self.offset else datetime.datetime.utcnow()
        )
        start_time: datetime.datetime = end_time - window_length
        super().__init__(start_time, end_time)



from unittest.mock import patch

import pytest
import responses

from gantry.alerts.client import AlertsAggregation, AlertsCheck, GantryAlerts
from gantry.api_client import APIClient
from gantry.exceptions import GantryException

from .conftest import ORIGIN


@pytest.fixture
def gantry_alerts_obj():
    api_client = APIClient(origin=ORIGIN, api_key="abcd1234")
    return GantryAlerts(api_client)


@pytest.mark.parametrize(
    ["data", "application_name", "expected"],
    [
        (
            {"data": {"id": "02d7c315-9083-476d-8fe6-4d41db6045b8"}, "response": "ok"},
            "test_app",
            {"id": "02d7c315-9083-476d-8fe6-4d41db6045b8"},
        )
    ],
)
def test_get_application(data, application_name, expected, gantry_alerts_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/models/{}/schemas".format(ORIGIN, application_name),
            json=data,
        )
        assert gantry_alerts_obj._get_application(application_name) == expected


@pytest.mark.parametrize(
    ["data", "application_name", "expected"],
    [
        (
            {"error": "Model not found", "response": "error"},
            "dne",
            None,
        ),
    ],
)
def test_get_application_for_nonexistent_app(data, application_name, expected, gantry_alerts_obj):
    with pytest.raises(GantryException):
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                "{}/api/v1/models/{}/schemas".format(
                    ORIGIN, application_name
                ),  # we mock the error response
                json=data,
                status=404,
            )
            assert gantry_alerts_obj._get_application(application_name) == expected


@patch(
    "gantry.alerts.client.GantryAlerts._get_application",
    return_value={
        "feedback_datanodes": [{"name": "one", "id": "5393dc5f-1304-4fc8-a06a-bba6f6483c1a"}],
        "prediction_datanodes": [{"name": "two", "id": "7382bc29-3ab1-4bc8-ac28-cd2a49886bdc"}],
        "projection_datanodes": [{"name": "three", "id": "7d4647fc-a447-4dfa-bf19-99e080404029"}],
    },
)
def test_map_column_names_to_ids(mock_req, gantry_alerts_obj):
    expected = {
        "one": "5393dc5f-1304-4fc8-a06a-bba6f6483c1a",
        "two": "7382bc29-3ab1-4bc8-ac28-cd2a49886bdc",
        "three": "7d4647fc-a447-4dfa-bf19-99e080404029",
    }
    actual = gantry_alerts_obj._map_column_names_to_ids("test app", ["one", "two", "three"])
    assert actual == expected


@patch(
    "gantry.alerts.client.GantryAlerts._get_application",
    return_value={},
)
def test_map_column_names_throws_when_node_not_found(mock_req, gantry_alerts_obj):
    with pytest.raises(ValueError):
        gantry_alerts_obj._map_column_names_to_ids("test app", ["one"])


@pytest.mark.parametrize(
    [
        "application_name",
        "triggered",
        "triggered_from",
        "triggered_to",
        "mock_resp_data",
    ],
    [
        (
            "test_app",
            False,
            None,
            None,
            [{"alert_id": "02342a0d-4bce-40e0-a8bd-82e5892ca2ea"}],
        )
    ],
)
def test_get_alerts(
    application_name,
    triggered,
    triggered_from,
    triggered_to,
    mock_resp_data,
    gantry_alerts_obj,
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/alerts".format(ORIGIN),
            json={"data": {"alerts": mock_resp_data}, "response": "ok"},
        )
        assert (
            gantry_alerts_obj.get_alerts(
                application_name=application_name,
                triggered=triggered,
                triggered_from=triggered_from,
                triggered_to=triggered_to,
            )
            == mock_resp_data
        )


@pytest.mark.parametrize(
    ["application_name", "triggered", "triggered_from", "triggered_to", "mock_resp_data"],
    [
        (
            "test_app",
            True,
            "1970-01-01T00:00:00",
            "2022-10-03T14:21:14.454125",
            [{"alert_id": "02342a0d-4bce-40e0-a8bd-82e5892ca2ea"}],
        ),
        (
            "test_app",
            True,
            None,
            None,
            [{"alert_id": "02342a0d-4bce-40e0-a8bd-82e5892ca2ea"}],
        ),
    ],
)
def test_get_triggered_alerts(
    application_name, triggered, triggered_from, triggered_to, mock_resp_data, gantry_alerts_obj
):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/alerts/triggered/all".format(ORIGIN),
            json={"data": {"triggered_alerts": mock_resp_data}, "response": "ok"},
        )
        assert (
            gantry_alerts_obj.get_alerts(
                application_name=application_name,
                triggered=triggered,
                triggered_from=triggered_from,
                triggered_to=triggered_to,
            )
            == mock_resp_data
        )


@pytest.mark.parametrize(
    ["application_name", "triggered_from", "triggered_to"],
    [
        ("test app", "2022-10-03T14:21:14.454125", None),
        ("test app", None, "2022-10-03T14:21:14.454125"),
    ],
)
def test_get_triggered_alerts_raises_if_not_filtering_and_specifiying_times(
    application_name, triggered_from, triggered_to, gantry_alerts_obj
):
    with pytest.raises(ValueError):
        gantry_alerts_obj.get_alerts(
            application_name=application_name,
            triggered=False,
            triggered_from=triggered_from,
            triggered_to=triggered_to,
        )


@patch(
    "gantry.api_client.APIClient.request",
    return_value={
        "data": {"alert": {"id": "3ca87947-5d26-41a9-9e9f-c07050188c6d"}},
        "response": "ok",
    },
)
@patch(
    "gantry.alerts.client.GantryAlerts._get_application",
    return_value={"id": "5bc22499-cb09-4984-a450-cf16c4a05c5f"},
)
@patch(
    "gantry.alerts.client.GantryAlerts._map_column_names_to_ids",
    return_value={"some.feature": "db144056-ec59-44ca-9ed5-4e8b162a4fbc"},
)
def test_create_alert(mock_map, mock_get_app, mock_req, gantry_alerts_obj):
    application_name = "test app"
    alert_name = "test alert"

    alert_checks = [
        AlertsCheck(
            **{
                "aggregation": AlertsAggregation.MAX,
                "column_names": ["some.feature"],
                "lower_bound": 1000,
                "upper_bound": 1001,
            }
        )
    ]
    evaluation_window = "PT1H5M26S"
    evaluation_delay = "PT1H5M26S"
    tags = {"env": "dev"}
    gantry_alerts_obj.create_alert(
        application_name=application_name,
        alert_name=alert_name,
        alert_checks=alert_checks,
        evaluation_window=evaluation_window,
        evaluation_delay=evaluation_delay,
        tags=tags,
    )
    mock_get_app.assert_called_once()
    mock_req.assert_called_with(
        "POST",
        "/api/v1/alerts",
        json={
            "alert_type": GantryAlerts.ALERT_TYPE,
            "alert_name": alert_name,
            "model_node_id": "5bc22499-cb09-4984-a450-cf16c4a05c5f",
            "relative_time": evaluation_window,
            "delay_time": evaluation_delay,
            "checks": [
                {
                    "aggregation": "maximum",
                    "lower_bound": 1000,
                    "upper_bound": 1001,
                    "check_type": "range_check",
                    "data_node_ids": ["db144056-ec59-44ca-9ed5-4e8b162a4fbc"],
                }
            ],
            "tags": tags,
        },
        raise_for_status=True,
    )


@patch(
    "gantry.api_client.APIClient.request",
    return_value={
        "data": {"alert": {"id": "3ca87947-5d26-41a9-9e9f-c07050188c6d"}},
        "response": "ok",
    },
)
@patch(
    "gantry.alerts.client.GantryAlerts._get_application",
    return_value={"id": "5bc22499-cb09-4984-a450-cf16c4a05c5f"},
)
@patch(
    "gantry.alerts.client.GantryAlerts._map_column_names_to_ids",
    return_value={"some.feature": "db144056-ec59-44ca-9ed5-4e8b162a4fbc"},
)
def test_update_alert(mock_map, mock_get_app, mock_req, gantry_alerts_obj):
    alert_id = "6848c6b3-e8bc-4d65-ba1c-7bace4350cbd"
    application_name = "test app"
    alert_name = "test alert"
    alert_checks = [
        AlertsCheck(
            **{
                "aggregation": AlertsAggregation.MAX,
                "column_names": ["some.feature"],
                "lower_bound": 1000,
                "upper_bound": 1001,
            }
        )
    ]
    evaluation_window = "PT1H5M26S"
    evaluation_delay = "PT1H5M26S"
    tags = {"env": "dev"}
    gantry_alerts_obj.update_alert(
        application_name=application_name,
        alert_id=alert_id,
        alert_name=alert_name,
        alert_checks=alert_checks,
        evaluation_window=evaluation_window,
        evaluation_delay=evaluation_delay,
        tags=tags,
    )
    mock_get_app.assert_called_once()
    mock_req.assert_called_with(
        "POST",
        "/api/v1/alerts",
        json={
            "alert_type": GantryAlerts.ALERT_TYPE,
            "alert_name": alert_name,
            "model_node_id": "5bc22499-cb09-4984-a450-cf16c4a05c5f",
            "alert_id": alert_id,
            "relative_time": evaluation_window,
            "delay_time": evaluation_delay,
            "checks": [
                {
                    "aggregation": "maximum",
                    "data_node_ids": ["db144056-ec59-44ca-9ed5-4e8b162a4fbc"],
                    "lower_bound": 1000,
                    "upper_bound": 1001,
                    "check_type": "range_check",
                }
            ],
            "tags": tags,
        },
        raise_for_status=True,
    )


@pytest.mark.parametrize(
    ["notify_daily", "daily_notification_time"],
    [
        (
            True,
            None,
        ),
        (True, "12:00:00"),
        (False, None),
    ],
)
@patch(
    "gantry.api_client.APIClient.request",
    return_value={
        "data": {"channel": {"id": "646e7bf1-fe23-4972-a5bf-896fd8f340c5"}},
        "response": "ok",
    },
)
def test_create_slack_notification(
    mock_req, notify_daily, daily_notification_time, gantry_alerts_obj
):
    gantry_alerts_obj.create_slack_notification(
        alert_id="6848c6b3-e8bc-4d65-ba1c-7bace4350cbd",
        notification_name="test notification",
        slack_webhook_url="test hook",
        notify_daily=notify_daily,
        daily_notification_time=daily_notification_time,
    )
    if daily_notification_time is not None:
        mock_req.assert_called_with(
            "POST",
            "/api/v1/alerts/notifications",
            json={
                "channel_type": GantryAlerts.CHANNEL_TYPE,
                "channel_name": "test notification",
                "webhook_url": "test hook",
                "alert_id": "6848c6b3-e8bc-4d65-ba1c-7bace4350cbd",
                "frequency": "DAILY" if notify_daily else "IMMEDIATELY",
                "frequency_relative_time": daily_notification_time,
            },
            raise_for_status=True,
        )
    else:
        mock_req.assert_called_with(
            "POST",
            "/api/v1/alerts/notifications",
            json={
                "channel_type": GantryAlerts.CHANNEL_TYPE,
                "channel_name": "test notification",
                "webhook_url": "test hook",
                "alert_id": "6848c6b3-e8bc-4d65-ba1c-7bace4350cbd",
                "frequency": "DAILY" if notify_daily else "IMMEDIATELY",
            },
            raise_for_status=True,
        )


@pytest.mark.parametrize(
    "mock_resp_data",
    [[{"id": "2d37db9b-fc01-41e2-9f20-4f31e74d78f3"}]],
)
def test_get_slack_notifications(mock_resp_data, gantry_alerts_obj):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.GET,
            "{}/api/v1/alerts/channels".format(ORIGIN),
            json={"data": {"channels": mock_resp_data}, "response": "ok"},
        )
        assert gantry_alerts_obj.get_slack_notifications() == mock_resp_data


@patch("gantry.api_client.APIClient.request", return_value={"response": "ok"})
def test_delete_alert(mock_req, gantry_alerts_obj):
    gantry_alerts_obj.delete_alert(alert_id="6848c6b3-e8bc-4d65-ba1c-7bace4350cbd")
    mock_req.assert_called_with(
        "DELETE", "/api/v1/alerts/6848c6b3-e8bc-4d65-ba1c-7bace4350cbd", raise_for_status=True
    )


@patch("gantry.api_client.APIClient.request", return_value={"response": "ok"})
def test_delete_slack_notification(mock_req, gantry_alerts_obj):
    gantry_alerts_obj.delete_slack_notification(
        alert_id="6848c6b3-e8bc-4d65-ba1c-7bace4350cbd",
        slack_notification_id="02342a0d-4bce-40e0-a8bd-82e5892ca2ea",
    )
    mock_req.assert_called_with(
        "DELETE",
        "/api/v1/alerts/channels",
        json={
            "alert_id": "6848c6b3-e8bc-4d65-ba1c-7bace4350cbd",
            "channel_id": "02342a0d-4bce-40e0-a8bd-82e5892ca2ea",
        },
        raise_for_status=True,
    )



import datetime
from typing import Optional, Union

from gantry.automations.curators.curators import Curator
from gantry.automations.curators.selectors import (
    TIME_COLUMN_NAME,
    BoundsFilter,
    EqualsFilter,
    OrderedSampler,
    OrderingDirection,
    Selector,
    StratificationMode,
    StratifiedSampler,
    UniformSampler,
)


class UniformCurator(Curator):
    """
    Stock Curator that uniformly samples application records and adds up to the
    top limit results to curated_dataset_name every curation_interval.
    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """

        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[Selector(method=UniformSampler(), limit=limit)],
        )


class NewestCurator(Curator):
    """
    Stock Curator that sorts the application records in descending time-order according
    to the internally logged Gantry timestamp, `__TIME`, and adds up to the top limit
    results to curated_dataset_name every curation_interval.
    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """

        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=OrderedSampler(
                        field=TIME_COLUMN_NAME,
                        sort=OrderingDirection.DESCENDING,
                    ),
                    limit=limit,
                )
            ],
        )


class OldestCurator(Curator):
    """
    Stock Curator that sorts the application records in ascending time-order according
    to the internally logged Gantry timestamp, `__TIME`, and adds up to the top limit
    results to curated_dataset_name every curation_interval.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """

        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=OrderedSampler(
                        field=TIME_COLUMN_NAME,
                        sort=OrderingDirection.ASCENDING,
                    ),
                    limit=limit,
                )
            ],
        )


class BalancedStratificationCurator(Curator):
    """
    Stock Curator that builds a balanced stratification of application records according
    to the field stratify_field and adds up to limit results to curated_dataset_name.

    Currently, this curator only supports stratification of categorical fields. And the limit
    per stratum is determineed by the stratum with the fewest records in the current
    curation_interval.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        stratify_field: str,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            stratify_field (str): The field to stratify.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=StratifiedSampler(
                        field=stratify_field,
                        mode=StratificationMode.balanced,
                    ),
                    limit=limit,
                ),
            ],
        )


class ProportionalStratificationCurator(Curator):
    """
    Stock Curator that builds a proportional stratification of application records
    according to the field stratify_field and adds up to limit results to
    curated_dataset_name.

    The proportion of records in each stratum is respected, but the limit per curation
    may be violated, unlike the StrictStratificationCurator, which will respect the limit
    at the (potential) cost of violating the proportion of records in each stratum.

    Additionally, it is worth noting that proportions are rounded to even numbers, so
    1.5 will be rounded to 2, and 0.5 will be rounded to 0.

    Example:
        Let's say there are 100 records: 51 records with the value "A" for the field, 15
        records with the value "B" for the field, and 16 records with the value "C" for the
        field, 17 records with the value "D" for the field, and 1 record with the value "E".

        If the limit is 10, then the curator will return 11 records, with 5 records with the
        value "A" for the field, 2 records with the value "B" for the field, 2 records with
        the value "C" for the field, 2 records with the value "D" for the field, and 0 records
        with the value "E".

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        stratify_field: str,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            stratify_field (str): The field to stratify.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=StratifiedSampler(
                        field=stratify_field,
                        mode=StratificationMode.proportional,
                    ),
                    limit=limit,
                ),
            ],
        )


class StrictStratificationCurator(Curator):
    """
    Stock Curator that builds a strict stratification of application records
    according to the field stratify_field and adds up to limit results to
    curated_dataset_name.

    The limit of records per curation is respected, but the proportion of records in each
    stratum may be violated, unlike the ProportionalStratificationCurator, which will
    respect the proportion at the (potential) cost of violating the limit per curation.

    Additionally, it is worth noting that the number of records to be added to each stratum
    is calulated exactly, leading to poentially fractional assignments. The integer part
    of the assignment is taken, and the total remainder is added, up to the limit, randomly.

    Example:
        Let's say there are 100 records: 51 records with the value "A" for the field, 15
        records with the value "B" for the field, and 16 records with the value "C" for the
        field, 17 records with the value "D" for the field, and 1 record with the value "E".

        If the limit is 10, then the curator will return 10 records, with 5 records with the
        value "A" for the field, 1 record with "B" for the field, 1 record with "C" for the
        field, 1 record with "D" for the field, 0 records with "E" for the field, and the
        remaining 2 records will be randomly selected from the remaining records.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        stratify_field: str,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            stratify_field (str): The field to stratify.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=StratifiedSampler(
                        field=stratify_field,
                        mode=StratificationMode.strict,
                    ),
                    limit=limit,
                ),
            ],
        )


class AscendedSortCurator(Curator):
    """
    Stock Curator that sorts the application field specified by sort_field in ascending order
    and adds up to the top limit results to curated_dataset_name every curation_interval.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        sort_field: str,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            sort_field (str): The field to sort.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=OrderedSampler(
                        field=sort_field,
                        sort=OrderingDirection.ASCENDING,
                    ),
                    limit=limit,
                ),
            ],
        )


class DescendedSortCurator(Curator):
    """
    Stock Curator that sorts the application field specified by sort_field in descending order
    and adds up to the top limit results to curated_dataset_name every curation_interval.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        sort_field: str,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            sort_field (str): The field to sort.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=OrderedSampler(
                        field=sort_field,
                        sort=OrderingDirection.DESCENDING,
                    ),
                    limit=limit,
                ),
            ],
        )


class BoundedRangeCurator(Curator):
    """
    Stock Curator that strictly bounds the application field specified by bound_field between
    upper_bound and lower_bound and adds up to limit results to curated_dataset_name
    every curation_interval. If the curation returns more results than limit, the results are
    sorted in ascending order by bound_field when sort_ascending is True (default) and in
    descending order when sort_ascending is False.

    """

    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        bound_field: str,
        upper_bound: float,
        lower_bound: float,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
        sort_ascending: bool = True,
    ):
        """
        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            bound_field (str): The field to bound.
            upper_bound (float): Upper bound on the field.
            lower_bound (float): Lower bound on the field.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
            sort_ascending (bool, optional): Whether or not to sort the results in ascending order.
                Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=OrderedSampler(
                        field=bound_field,
                        sort=OrderingDirection.ASCENDING
                        if sort_ascending
                        else OrderingDirection.DESCENDING,
                    ),
                    limit=limit,
                    filters=[BoundsFilter(field=bound_field, upper=upper_bound, lower=lower_bound)],
                ),
            ],
        )


class EqualityCurator(Curator):
    def __init__(
        self,
        name: str,
        application_name: str,
        limit: int,
        field: str,
        equals: Union[bool, str, int, float],
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curate_past_intervals: bool = True,
    ):
        """
        Stock Curator that selects all results where the field specified by field is equal to
        equals and adds up to limit results to curated_dataset_name every curation_interval.

        If the curation returns more results than limit, the results are sampled uniformly at
        random.

        Args:
            name (str): The name of the curator.
            application_name (str): The name of the application that the curator is running in.
            limit (int): The maximum number of results to return.
            field (str): The field to equate.
            equals (Union[bool, str, int, float]): The value to equate the field to. String values
                "true" and "false" are validated as bool types.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset that
                the curator will write to. Defaults to None, in which case the name of the curator
                will be used.
            start_on (Optional[datetime.datetime], optional): The datetime to start the
                curator. Defaults to None, in which case the curator will start now, looking back
                one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval of time considered
            during each run of the curator. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether or not to curate past intervals when
            start_on is in the past. Defaults to True.
        """
        super().__init__(
            name=name,
            curated_dataset_name=curated_dataset_name,
            application_name=application_name,
            start_on=start_on,
            curation_interval=curation_interval,
            curate_past_intervals=curate_past_intervals,
            selectors=[
                Selector(
                    method=UniformSampler(),
                    limit=limit,
                    filters=[EqualsFilter(field=field, equals=equals)],
                ),
            ],
        )



import datetime

CURRENT_TIME, CURRENT_TIME_STR = datetime.datetime(2032, 5, 23, 0, 0, 0), "2032-05-23T00:00:00"
FUTURE_TIME = datetime.datetime(2042, 5, 23, 0, 0, 0)
SOME_TIME, SOME_TIME_STR = datetime.datetime(2022, 5, 10, 17, 50, 32), "2022-05-10T17:50:32"
ANOTHER_TIME, ANOTHER_TIME_STR = datetime.datetime(2022, 5, 11, 17, 50, 32), "2022-05-11T17:50:32"
HOST = "http://test-api"
ONE_TO_TEN = list(range(1, 10))



from distutils.util import convert_path

import setuptools

version = {}
with open(convert_path("gantry/version.py")) as fp:
    exec(fp.read(), version)


install_requires = [
    "datasets~=2.7.0",
    "Jinja2~=3.1.2",
    "aiohttp~=3.8",
    # 1.1.1 contains a fix for a regression in a dependency
    # https://stackoverflow.com/a/71574145
    "dateparser~=1.1,>=1.1.1",
    "typeguard~=2.13",
    "boto3-extensions>=0.11.0",
    "backoff~=2.1.2",
    "cachetools~=4.2",
    "click>=7.1.2",
    "click-aliases~=1.0",
    "click-spinner>=0.1.10",
    "requests>=2.22.0",
    "pyyaml>=5.3.1",
    "colorama>=0.4.0",
    "label-studio-sdk>=0.0.19,<0.0.22",
    "marshmallow>=3.11.1",
    "marshmallow-oneofschema>=3.0.1",
    "monotonic>=1.4",
    "tabulate~=0.8",
    "typing-extensions>=4.6",
    "tqdm~=4.64",
    "halo~=0.0.31",
    "isodate~=0.6",
    "pydantic>=1.10.8",
    "python-dateutil>=2.8.1",
    "pandas",
    "numpy",
]


# https://packaging.python.org/tutorials/packaging-projects/
setuptools.setup(
    name="gantry",
    version=version["__version__"],
    install_requires=install_requires,
    include_package_data=True,
    author="Gantry Systems, Inc.",
    author_email="oss@gantry.io",
    description="Gantry Python Library",
    long_description="",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license="Apache Software License v2",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": ["gantry-cli = gantry.cli.main:cli"],
    },
    package_data={"": ["dataset/templates/*.jinja"]},
)



import uuid

import responses
from responses import matchers

from .conftest import HOST


def test_create_application(test_client):
    test_id = uuid.uuid4()
    application_name = "test_application_name"
    application_type = "test_application_type"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/applications",
            json={
                "response": "ok",
                "data": {
                    "id": str(test_id),
                    "model_name": application_name,
                },
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {"model_name": application_name, "model_type": application_type}
                )
            ],
        )
        test_application = test_client.create_application(application_name, application_type)

    assert test_application._name == application_name
    assert test_application._id == test_id


def test_get_application(test_client):
    test_id = uuid.uuid4()
    application_name = "test_application_name"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application_name}",
            json={
                "response": "ok",
                "data": {
                    "versions": [
                        {
                            "internal_version": "1",
                            "version": "first model",
                            "application_id": str(test_id),
                        },
                    ],
                    "func_name": application_name,
                },
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application_name}/schemas",
            json={
                "response": "ok",
                "data": {
                    "model_type": "test_model_type",
                },
            },
        )
        test_application = test_client.get_application(application_name)

    assert test_application._name == application_name
    assert test_application._id == test_id
